#lang racket
(require racket/runtime-path
         rackunit
         "../git.rkt")

(define-runtime-module-path-index main "../main.rkt")

(define work-dir (make-temporary-file "~a-git-slice"
                                      'directory))

(define (reset-dir dir)
  (delete-directory/files #:must-exist? #f dir)
  (make-directory* dir))

(define repo-dir (build-path work-dir "repo"))
(reset-dir repo-dir)

(define commit-counts (hash))

;; ----------------------------------------
;; Script a  repository with interesting moves, branching, etc.

(parameterize ([current-directory repo-dir])
  (system "git init")
  
  (define (slices->msg line l)
    (~s (apply ~a #:separator " " "CHANGED:" line l)))
  (define (merge-counts a b)
    (for/fold ([a a]) ([(k v) (in-hash b)])
      (hash-update a k (lambda (n) (+ n v)) 0)))
  (define (increment-counts counts slices)
    (for/fold ([ht counts]) ([slice (in-list slices)])
      (hash-update ht slice add1 0)))
  
  (define n 32)
  (define (create p)
    ;; Create a file that will not be considered a copy
    ;; of any other file.
    (make-directory* (path-only p))
    (call-with-output-file*
     p
     (lambda (o)
       (for ([i 100])
         (display (random 100) o)
         (displayln (make-bytes n 60) o))
       (set! n (add1 n)))))
  (define (modify p)
    (define s (file->bytes p))
    (call-with-output-file*
     #:exists 'update
     p
     (lambda (o)
       (displayln (bytes-append
                   (subbytes s 0 n)
                   (make-bytes 100 n)
                   (subbytes s n))
                  o)
       (set! n (add1 n)))))
  (define (move s d)
    (make-directory* (path-only d))
    (system (~a "git mv " s " " d)))
  (define (copy s d)
    (make-directory* (path-only d))
    (copy-file s d))
  (define (do-commit line . slices)
    (set! commit-counts (increment-counts commit-counts slices))
    (system (~a "git add . && git commit -m " (slices->msg line slices))))
  (define-syntax (commit stx)
    (syntax-case stx ()
      [(_ slice ...)
       #`(do-commit #,(syntax-line stx) slice ...)]))
  (define (fork a b)
    (define old-counts commit-counts)
    (set! commit-counts (hash))
    (system "git branch left && git checkout left")
    (a)
    (define left-counts commit-counts)
    (set! commit-counts (hash))
    (system "git checkout master")
    (b)
    (define right-counts commit-counts)
    (define both-counts (merge-counts left-counts right-counts))
    (define slices (hash-keys both-counts))
    (system (~a "git merge -m " (slices->msg 0 slices) " left"
                " && git branch -d left"))
    (set! commit-counts (merge-counts old-counts both-counts)))
  
  ;; ------------------------------------------------------------
  ;; Files with names ending in "..._Alpha" will end up in "alpha",
  ;; etc., and each of those directories corresponds to a slice to
  ;; try. Each `commit` names the end slices that should include
  ;; the commit.
  ;;
  ;; Gamma is created in one branch, and Delta in another
  ;; Epsilon is created in two branches, which would give it
  ;;  multiple initial commits without special handling
  ;; Zeta has modifications in two branches, and none
  ;;  after the branch
  
  (create "a/x_Alpha")
  (commit "Alpha")
  
  (modify "a/x_Alpha")
  (create "a/y_Alpha")
  (commit "Alpha")
  
  (create "b/x_Beta")
  (create "b/y_Beta")
  (commit "Beta")
  
  (move "a/x_Alpha" "a/z_Alpha")
  (move "a/y_Alpha" "c/y_Alpha")
  (commit "Alpha")
  
  (copy "b/x_Beta" "b/z_Beta")
  (commit "Beta")
  
  (create "z/x_Zeta")
  (create "z/y_Zeta")
  (commit "Zeta")
  
  (fork
   (lambda ()
     (create "c/x_Gamma")
     (commit "Gamma")
     
     (modify "a/z_Alpha")
     (commit "Alpha")
     
     (create "e/x_Epsilon")
     (commit "Epsilon")
     
     (move "z/x_Zeta" "zeta/x_Zeta")
     (commit "Zeta"))
   (lambda ()
     (create "d/x_Delta")
     (commit "Delta")
     
     (modify "b/z_Beta")
     (move "b/y_Beta" "c/y_Beta")
     (commit "Beta")
     
     (create "e/y_Epsilon")
     (commit "Epsilon")
     
     (move "z/y_Zeta" "zeta/y_Zeta")
     (commit "Zeta")))
  
  ;; Move all into place:
  (move "a/z_Alpha" "alpha/x_Alpha")
  (move "c/y_Alpha" "alpha/y_Alpha")
  (move "b/x_Beta" "beta/x_Beta")
  (move "c/y_Beta" "beta/y_Beta")
  (move "b/z_Beta" "beta/z_Beta")
  (move "c/x_Gamma" "gamma/x_Gamma")
  (move "d/x_Delta" "delta/x_Delta")
  (move "e/x_Epsilon" "epsilon/x_Epsilon")
  (move "e/y_Epsilon" "epsilon/y_Epsilon")
  (commit "Alpha" "Beta" "Gamma" "Delta" "Epsilon"))

;; ----------------------------------------
;; Extract and check slices

(define (slice slice [extra-commit? #f])
  (printf "~a\nSLICING ~a...\n" (make-string 60 #\=) slice)
  (define slice-dir (build-path work-dir "slice"))
  (reset-dir slice-dir)
  (parameterize ([current-directory work-dir])
    (system "git clone repo slice"))
  
  (define dir (string-foldcase slice))
  
  (parameterize ([current-directory slice-dir]
                 [current-command-line-arguments (vector dir)]
                 [current-namespace (make-base-namespace)])
    (define-values (name base) (module-path-index-split main))
    (dynamic-require (module-path-index-join name base) #f))
  
  (check-equal? (directory-list slice-dir)
                (map string->path (list ".git" dir)))
  
  (parameterize ([current-directory slice-dir])
    (define o (open-output-bytes))
    (parameterize ([current-output-port o])
      (system "git log"))
    (define commit-count
      (for/fold ([count 0]) ([l (in-lines (open-input-string
                                           (get-output-string o)))])
        (cond
         [(regexp-match? #rx"^commit " l)
          (add1 count)]
         [(regexp-match? #rx"^CHANGED: " l)
          (check-true (member slice (string-split l)))]
         [else count])))
    (define expected-count  (+ (hash-ref commit-counts slice)
                               (if extra-commit? 1 0)))
    (check-equal? commit-count expected-count)
    (unless (equal? commit-count expected-count)
      (exit))
    
    ;; Checks that we didn't create a repo with multiple
    ;; initial commits (which is a danger with Epsilon):
    (call-with-values extract-commits void)))

(slice "Alpha")
(slice "Beta")
(slice "Gamma")
(slice "Delta")
(slice "Epsilon" #t) ; extra commit is a new initial commit
(slice "Zeta" #t) ; extra commit is a new head to join branches

;; ----------------------------------------

(delete-directory/files work-dir)
